%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Main Script for Japanese Vowels Task %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This script implements the "Japanese Vowels" speaker classification
% benchmark, as documented in the paper: Herbert Jaeger, Mantas
% Lukosevicius, Dan Popovici, Optimization and Applications of Echo State
% Networks with Leaky Integrator Neurons. Neural Networks, April 2007 
% (preprint  available from first author's publications page)

% created and copyright: Herbert Jaeger, Feb 23, 2007

% This software is free for non-commercial use. If commercial use is
% intended, contact Fraunhofer IAIS (www.iais.fraunhofer.de) who have
% claimed international patents on ESN algorithms (pending).

% This software is intended for research use by experienced Matlab users 
% and includes no warranties or services. 
% Bug reports are gratefully received by h.jaeger@iu-bremen.de

addpath ./tools;
addpath ./data;

%%%%%%%%%%%%% Setting up the number of expert (voters) networks 

newNets = 1;  % Set to 1 if new networks are to be created, else 0
nExperts = 200; % Number of voting experts. Note: the script is quite slow, on 
                % a 300 MHz PC 100 voters take about 1 hour.

nInternalUnits = 4; % reservoir size per voter, can be adjusted
nInputUnits = 14; % number of input units, fixed at 14 for this task
nOutputUnits = 9; % number of output units, fixed at 9 for this task

%%%%%%%%%%%% Data preparation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
JapVouwels_dataPrep; % A script that reads in the original data 
                     % (which come in text format) and assigns them 
                     % to trainInputs, testInputs, trainOutputs,
                     % testOutputs
[trainInputs,  shifts] = normalizeData_min0(trainInputs); % shift trainInputs
                     % signals such that global min per channel is 0, and
                     % also return the 14 shifts
testInputs = shiftData(testInputs,  shifts); % submit the test Inputs to 
                     % the same shifts

%%%%%%%%%% Initialization %%%%%%%%%%%%%%%%%%%%%%%%%%  

%%% allocating memory for the expert networks' outputs
trainOutsIndividual = zeros(270, 9, nExperts); 
testOutsIndividual = zeros(370, 9, nExperts);

%%% Create new ESNs (called "experts" here). 
if newNets
    experts = cell(nExperts, 1);
    for n = 1:nExperts
        experts{n,1} = generate_esn(nInputUnits, nInternalUnits, nOutputUnits, ...
            'spectralRadius', 0.2, 'inputScaling', ones(14,1), 'inputShift', zeros(14,1), ...
            'teacherScaling', 0.8 * ones(9,1), 'teacherShift', zeros(9,1), 'feedbackScaling', 0,...
            'methodWeightCompute', 'wiener_hopf', 'learningMode', 'offline_multipleTimeSeries',...
            'type', 'leaky_esn'); 
    end
end


%%%%%%%%%%%% training and testing. This is done in a loop over the expert
%%%%%%%%%%%% networks.

for n = 1:nExperts
    disp(n);
    %%% (Possibly re-)set the expert control/scaling parameters 
    experts{n,1}.spectralRadius = 0.2;
    experts{n,1}.internalWeights = experts{n,1}.internalWeights_UnitSR * experts{n,1}.spectralRadius;
    experts{n,1}.outputActivationFunction = 'tanh';
    experts{n,1}.inverseOutputActivationFunction = 'atanh';
    experts{n,1}.teacherScaling = 1.8;
    experts{n,1}.teacherShift = -0.9 * ones(9,1);
    inScFac = 1.5;
    experts{n,1}.inputScaling = inScFac * ones(14,1); experts{n,1}.inputScaling(end-1:end,1) = 0.1;
    inShFac = 0;
    experts{n,1}.inputShift = inShFac * ones(14,1);
    experts{n,1}.noiseLevel = 0.0;
    experts{n,1}.leakage = 0.2;
    experts{n,1}.timeConstant = 1;
    experts{n,1}.retainment = 1 - experts{n,1}.leakage * experts{n,1}.timeConstant;
    experts{n,1}.suppNr = 3 ;
    experts{n,1}.methodWeightCompute = 'pseudoinverse';
    experts{n,1}.type = 'leaky_esn'; 
    
    experts{n,1}.outputWeights = ...
        zeros(experts{n,1}.nOutputUnits, ...
        experts{n,1}.suppNr * (experts{n,1}.nInternalUnits + experts{n,1}.nInputUnits));
    
     
    %%% some diagnostic / bookkeeping textual output at beginning of
    %%% experiment
    if n == 1 
        disp(sprintf('N %g  SR %g  inScale %g  inShift %g  NL %g  suppNR %g  leakage %g   %s', ...
            experts{n,1}.nInternalUnits, experts{n,1}.spectralRadius, inScFac, ...
            inShFac, experts{n,1}.noiseLevel, experts{n,1}.suppNr, ...
            experts{n,1}.leakage, experts{n,1}.type));        
        
            experts{n,1}.retainment = 1 - experts{n,1}.leakage * experts{n,1}.timeConstant;
            checkMat = experts{n,1}.internalWeights + experts{n,1}.retainment * ...
                eye(experts{n,1}.nInternalUnits);
            checkMat = experts{n,1}.internalWeights + experts{n,1}.retainment * ...
                eye(experts{n,1}.nInternalUnits);
            opts.disp = 0 ; 
            effective_specRad = max(abs(eigs(checkMat,1,'LM',opts)));
            disp(sprintf('eff. SR of leaky ESN: %g', effective_specRad));
    end
    
    
    
    %%%%%%%%%%%% Training %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % trainedESN is the ESN after training, that is, it is identical to
    % experts{n,1} except that the output weights trainedESN.outputWeights
    % are now trained. The other output, stateCollection, contains the
    % harvested network states, but this is not used in this script.
    [trainedESN stateCollection] = ...
        train_esn_JapVowels(trainInputs, trainOutputs, experts{n,1}) ; 
    experts{n,1} = trainedESN;
    
    %%%%%%%%%%%% Testing %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    %%%% compute outputs of trained ESN on training data
    
    % trainOutsIndividual(i,:,n) is the 9-vote vector of the n-th network
    % on the i-th training input
    for i = 1:270    
        trainOutsIndividual(i,:,n) = ...
            test_esn_JapVow(trainInputs{i,1}, trainedESN);   
    end
    
    %%%% compute outputs of trained ESN on testing data
    for i = 1:370    
        testOutsIndividual(i,:,n) = ...
            test_esn_JapVow(testInputs{i,1}, trainedESN);    
    end
end

%%% run the script that computes diagnostics
JapVow_DiagScript;