function [trained_esn, stateCollection] = ...
    train_esn_JapVowels(trainInput, trainOutput , esn)
% A specialized version of ESN training for the situation in the Japanese 
% Vowels experiment. To be used in the context of the script
% "JapVow_MainScript" in the demo package associated with the paper: Herbert Jaeger, Mantas
% Lukosevicius, Dan Popovici, Optimization and Applications of Echo State
% Networks with Leaky Integrator Neurons. Neural Networks, April 2007 
% (preprint  available from first author's publications page)

% Adapted from train_esn (Version 1.0, April 30, 2006) June 30, 2006, H. Jaeger. 
% For terms of use, see JapVow_MainScript.

% inputs:
% trainInput: input vector of size nTrainingPoints x nInputDimension
% trainOutput:  teacher vector of size nTrainingPoints x nOutputDimension
% esn:  an ESN structure, through which we run our input sequence

% outputs: 
% trained_esn:  an Esn structure with the option trained = 1 and 
% outputWeights set. 
% stateCollection:  matrix of size (nTrainingPoints-nForgetPoints) x
% nInputUnits + nInternalUnits 
% stateCollectMat(i,j):  internal activation of unit j after the 
% (i + nForgetPoints)th training point has been presented to the network
% teacherCollection is a nSamplePoints * nOuputUnits matrix that keeps
% the expected output of the ESN
% teacherCollection is the transformed(scaled, shifted etc) output see
% compute_teacher for more documentation



trained_esn = esn;
if not(strcmp(trained_esn.learningMode, 'offline_multipleTimeSeries'))
    error('learningMode must be set to offline_multipleTimeSeries in this Japanese Vowel script');
else
    
    % trainInput and trainOutput each represent a collection of K time
    % series, given in cell arrays of size K x 1, where each cell is an
    % array of size individualSequenceLength x sequenceDimension
    
    nrSamples = size(trainInput,1);
    
    % collect input+reservoir states into stateCollection
    stateCollection = zeros(nrSamples, ...
        esn.suppNr * (trained_esn.nInputUnits + trained_esn.nInternalUnits));
    
    for i = 1:nrSamples 
        stateCollection_i = ...
            compute_statematrix(trainInput{i,1}, trainOutput{i,1}, trained_esn, 0);
        l = size(trainInput{i,1},1);
        % boil states down to only suppNr many; use interpolation if
        % necessary
        for s = 1:esn.suppNr
            if s < esn.suppNr
                lowerInd = floor(s * (l-1)/esn.suppNr) + 1;
                a = s * (l-1)/esn.suppNr - floor(s * (l-1)/esn.suppNr);
                addState = (1 - a) * stateCollection_i(lowerInd,:) + ...
                    a * stateCollection_i(lowerInd + 1,:) ;
            else
                addState = stateCollection_i(end,:);
            end
            stateCollection(i, ...
                (s-1)*(trained_esn.nInputUnits + trained_esn.nInternalUnits)+1:...
                s * (trained_esn.nInputUnits + trained_esn.nInternalUnits),:) = addState;                
        end        
    end
    
    % collect teacher signals (including applying the inverse output
    % activation function) into teacherCollection
    teacherCollection = zeros(nrSamples, trained_esn.nOutputUnits);
    for i = 1:nrSamples            
        teacherCollection(i, :) = ...
            compute_teacher(trainOutput{i,1}(1, :), trained_esn, 0);             
    end
    
    % compute output weights
    
    trained_esn.outputWeights = ...
        feval(trained_esn.methodWeightCompute, stateCollection, teacherCollection) ;
end       


trained_esn.trained = 1 ;          


